from selve.util import *


class DeviceScanStart(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoDeviceCommand.SCANSTART)


class DeviceScanStop(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoDeviceCommand.SCANSTOP)


class DeviceSave(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoDeviceCommand.SAVE, [(ParameterType.INT, id)])


class DeviceGetIds(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoDeviceCommand.GETIDS)


class DeviceGetInfo(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoDeviceCommand.GETINFO, [(ParameterType.INT, id)])


class DeviceGetValues(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoDeviceCommand.GETVALUES, [(ParameterType.INT, id)])


class DeviceSetFunction(GatewayCommand):
    def __init__(self, id: int, function: DeviceFunctions):
        super().__init__(CommeoDeviceCommand.SETFUNCTION,
                         [(ParameterType.INT, id), (ParameterType.INT, function.value)])


class DeviceSetLabel(GatewayCommand):
    def __init__(self, id: bool, name: str):
        super().__init__(CommeoDeviceCommand.SETLABEL, [(ParameterType.INT, id), (ParameterType.STRING, name)])


class DeviceSetType(GatewayCommand):
    def __init__(self, id: int, type: DeviceType):
        super().__init__(CommeoDeviceCommand.SETTYPE, [(ParameterType.INT, id), (ParameterType.INT, type.value)])


class DeviceDelete(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoDeviceCommand.DELETE, [(ParameterType.INT, id)])


class DeviceWriteManual(GatewayCommand):
    def __init__(self, id: int, adress: int, name: str, type: DeviceType):
        super().__init__(CommeoDeviceCommand.WRITEMANUAL,
                         [(ParameterType.INT, id), (ParameterType.INT, adress), (ParameterType.STRING, name),
                          (ParameterType.INT, type.value)])


class DeviceScanStartResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class DeviceScanStopResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class DeviceScanResultResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.scanState = ScanState(int(parameters[0][1]))
        self.noNewDevices = int(parameters[1][1])
        self.foundIds = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[2][1]))]


class DeviceSaveResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class DeviceGetIdsResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ids = [b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[0][1]))]


class DeviceGetInfoResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = parameters[0][1]
        self.rfAddress = parameters[2][1]
        self.deviceType = DeviceType(int(parameters[3][1]))
        self.state = DeviceState(int(parameters[4][1]))


class DeviceGetValuesResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = parameters[0][1] if parameters[0][1] else ""
        self.movementState = MovementState(int(parameters[2][1])) if int(parameters[2][1]) else MovementState(0)
        self.value = Util.valueToPercentage(int(parameters[3][1]))
        self.targetValue = Util.valueToPercentage(int(parameters[4][1]))

        bArr = Util.intToBoolarray(int(parameters[5][1]))
        self.unreachable = bArr[0]
        self.overload = bArr[1]
        self.obstructed = bArr[2]
        self.alarm = bArr[3]
        self.lostSensor = bArr[4]
        self.automaticMode = bArr[5]
        self.gatewayNotLearned = bArr[6]
        self.windAlarm = bArr[7]
        self.rainAlarm = bArr[8]
        self.freezingAlarm = bArr[9]

        self.dayMode = DayMode(int(parameters[6][1]))


class DeviceSetFunctionResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class DeviceSetLabelResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class DeviceSetTypeResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class DeviceDeleteResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class DeviceWriteManualResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])



from selve import SelveTypes, Util, DeviceType, CommunicationType, MovementState, DayMode, \
    DeviceGetIdsResponse, DeviceGetInfoResponse, DeviceGetInfo, DeviceGetValuesResponse, DeviceGetValues


class SelveDevice:
    def __init__(self, id: int, device_type: SelveTypes = SelveTypes.UNKNOWN,
                 device_sub_type: DeviceType = DeviceType.UNKNOWN):
        self.id = id
        self.device_type = device_type
        self.device_sub_type = device_sub_type
        self.mask = Util.singlemask(id)
        self.name = "None"
        self.rfAdress = 0
        self.communicationType = CommunicationType.COMMEO
        self.state = MovementState.UNKOWN
        self.infoState = 0
        self.value = 0
        self.targetValue = 0
        self.unreachable = False
        self.overload = False
        self.obstructed = False
        self.alarm = False
        self.lostSensor = False
        self.automaticMode = False
        self.gatewayNotLearned = False
        self.windAlarm = False
        self.rainAlarm = False
        self.freezingAlarm = False
        self.dayMode = DayMode.UNKOWN


    def __str__(self):
        return "Device " + self.device_type.name + " " + self.device_sub_type.name + " of type: " + self.communicationType.name + " on channel " + str(
            self.id) + " with name " + self.name








from selve import SelveTypes, DeviceType, Util, CommunicationType, windDigital, rainDigital, tempDigital, lightDigital, \
    SensorState


class SelveSenSim:
    def __init__(self, id: int, device_type: SelveTypes = SelveTypes.SENSIM,
                 device_sub_type: DeviceType = DeviceType.UNKNOWN):
        self.id = id
        self.device_type = device_type
        self.device_sub_type = device_sub_type
        self.mask = Util.singlemask(id)
        self.name = "SensorSim " + str(id)
        self.activity = ""
        self.communicationType = CommunicationType.COMMEO
        self.windDigital = windDigital.NONE
        self.rainDigital = rainDigital.NONE
        self.tempDigital = tempDigital.NONE
        self.lightDigital = lightDigital.NONE
        self.sensorState = SensorState.AVAILABLE
        self.tempAnalog = 0
        self.windAnalog = 0
        self.sun1Analog = 0
        self.dayLightAnalog = 0
        self.sun2Analog = 0
        self.sun3Analog = 0
    def __str__(self):
        return "Sensor " + self.device_type.name + " of type: " + self.communicationType.name + " on channel " + str(
            self.id) + " with name " + self.name



from selve.util import *


class ParamSetForward(GatewayCommand):
    def __init__(self, on: bool):
        super().__init__(CommeoParamCommand.SETFORWARD, [(ParameterType.INT, 1 if on else 0)])


class ParamGetForward(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoParamCommand.GETFORWARD)


class ParamSetEvent(GatewayCommand):
    def __init__(self, EventDevice: bool, EventSensor: bool, EventSender: bool, Logging: bool, EventDuty: bool):
        super().__init__(CommeoParamCommand.SETEVENT,
                         [(ParameterType.INT, EventDevice), (ParameterType.INT, EventSensor),
                          (ParameterType.INT, EventSender), (ParameterType.INT, Logging),
                          (ParameterType.INT, EventDuty)])


class ParamGetEvent(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoParamCommand.GETEVENT)


class ParamGetDuty(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoParamCommand.GETDUTY)


class ParamGetRf(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoParamCommand.GETRF)


class ParamSetForwardResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class ParamGetForwardResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.forwarding = Forwarding(int(parameters[0][1]))


class ParamSetEventResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class ParamGetEventResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.eventDevice = bool(parameters[0][1])
        self.eventSensor = bool(parameters[1][1])
        self.eventSender = bool(parameters[2][1])
        self.eventLogging = bool(parameters[3][1])
        self.eventDuty = bool(parameters[4][1])


class ParamGetDutyResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.dutyMode = DutyMode(int(parameters[0][1]))
        self.rfTraffic = int(parameters[1][1])


class ParamGetRfResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.netAddress = int(parameters[0][1])
        self.resetCount = int(parameters[1][1])
        self.rfBaseId = int(parameters[2][1])
        self.sensorNetAddress = int(parameters[3][1])
        self.rfSensorId = int(parameters[4][1])
        self.iveoResetCount = int(parameters[5][1])
        self.rfIveoId = int(parameters[6][1])



from selve.util import *


class SensorTechStart(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSensorCommand.TEACHSTART)


class SensorTeachStop(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSensorCommand.TEACHSTOP)


class SensorGetIds(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSensorCommand.GETIDS)


class SensorGetInfo(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSensorCommand.GETINFO, [(ParameterType.INT, id)])


class SensorGetValues(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSensorCommand.GETVALUES, [(ParameterType.INT, id)])


class SensorSetLabel(GatewayCommand):
    def __init__(self, id: bool, name: str):
        super().__init__(CommeoSensorCommand.SETLABEL, [(ParameterType.INT, id), (ParameterType.STRING, name)])


class SensorDelete(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSensorCommand.DELETE, [(ParameterType.INT, id)])


class SensorWriteManual(GatewayCommand):
    def __init__(self, id: int, adress: int, name: str):
        super().__init__(CommeoSensorCommand.WRITEMANUAL,
                         [(ParameterType.INT, id), (ParameterType.INT, adress), (ParameterType.STRING, name)])


class SensorTeachStartResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SensorTeachStopResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SensorTeachResultResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.teachState = TeachState(int(parameters[0][1]))
        self.timeLeft = int(parameters[1][1])
        self.foundId = int(parameters[2][1])


class SensorGetIdsResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ids = [b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[0][1]))]


class SensorGetInfoResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = parameters[0][1]
        self.rfAddress = parameters[2][1]


class SensorGetValuesResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.windDigital = windDigital(int(parameters[1][1]))
        self.rainDigital = rainDigital(int(parameters[2][1]))
        self.tempDigital = tempDigital(int(parameters[3][1]))
        self.lightDigital = lightDigital(int(parameters[4][1]))
        self.sensorState = SensorState(int(parameters[5][1]))
        self.tempAnalog = int(parameters[6][1])
        self.windAnalog = int(parameters[7][1])
        self.sun1Analog = int(parameters[8][1])
        self.dayLightAnalog = int(parameters[9][1])
        self.sun2Analog = int(parameters[10][1])
        self.sun3Analog = int(parameters[11][1])


class SensorSetLabelResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SensorDeleteResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SensorWriteManualResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])



from selve.util import *


class GroupRead(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoGroupCommand.READ, [(ParameterType.INT, id)])


class GroupWrite(GatewayCommand):
    def __init__(self, id: int, actorIds: dict, name: str):
        super().__init__(CommeoGroupCommand.WRITE,
                         [(ParameterType.INT, id), (ParameterType.BASE64, Util.multimask(actorIds)),
                          (ParameterType.STRING, name)])


class GroupGetIds(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoGroupCommand.GETIDS)


class GroupDelete(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoGroupCommand.DELETE, [(ParameterType.INT, id)])


class GroupReadResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.id = int(parameters[1][1])
        self.mask = parameters[2][1]
        self.name = str(parameters[0][1])


class GroupWriteResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class GroupGetIdsResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ids = [b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[0][1]))]


class GroupDeleteResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])



from selve.util import *
        
class SenderTeachStart(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSenderCommand.TEACHSTART)
        
class SenderTeachStop(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSenderCommand.TEACHSTOP)
        
class SenderTeachResult(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSenderCommand.TEACHRESULT)
        
class SenderGetIds(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSenderCommand.GETIDS)
        
class SenderGetInfo(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenderCommand.GETINFO, [(ParameterType.INT, id)])
        
class SenderGetValues(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenderCommand.GETVALUES, [(ParameterType.INT, id)])
        
class SenderSetLabel(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenderCommand.SETLABEL, [(ParameterType.INT, id)])
        
class SenderDelete(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenderCommand.DELETE, [(ParameterType.INT, id)])
        
class SenderWriteManual(GatewayCommand):
    def __init__(self, id: int, adress: int, channel: int, resetCount: int, name: str):
        super().__init__(CommeoSenderCommand.WRITEMANUAL, [(ParameterType.INT, id), (ParameterType.INT, adress), (ParameterType.INT, channel), (ParameterType.INT, resetCount), (ParameterType.STRING, name)])




class SenderTeachStartResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class SenderTeachStopResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class SenderTeachResultResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = str(parameters[0][1])
        self.teachState = TeachState(int(parameters[1][1]))
        self.timeLeft = int(parameters[1][1])
        self.senderId = int(parameters[2][1])
        self.senderEvent = senderEvents(int(parameters[3][1]))

class SenderGetIdsResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ids = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[0][1]))]

class SenderGetInfoResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = parameters[0][1] if parameters[0][1] else ""
        self.rfAddress = parameters[2][1]
        self.rfChannel = int(parameters[3][1])
        self.rfResetCount = int(parameters[4][1])

class SenderGetValuesResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.lastEvent = senderEvents(int(parameters[1][1]))

class SenderSetLabelResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class SenderDeleteResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class SenderWriteManualResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])



from selve import MovementState, CommunicationType, Util, DeviceType, SelveTypes


class IveoDevice:
    def __init__(self, id: int, device_type: SelveTypes = SelveTypes.IVEO,
                 device_sub_type: DeviceType = DeviceType.UNKNOWN):
        self.id = id
        self.device_type = device_type
        self.device_sub_type = device_sub_type
        self.mask = Util.singlemask(id)
        self.name = "None"
        self.rfAdress = 0
        self.communicationType = CommunicationType.IVEO
        self.state = MovementState.UNKOWN
        self.activity = 0
        self.value = 0
        self.targetValue = 0


    def __str__(self):
        return "Device " + self.device_type.name + " of type: " + self.communicationType.name + " on channel " + str(
            self.id) + " with name " + self.name





from selve.util import *


class SenSimStore(GatewayCommand):
    def __init__(self, id: int, actorId: int):
        super().__init__(CommeoSenSimCommand.STORE, [(ParameterType.INT, id), (ParameterType.INT, actorId)])


class SenSimDelete(GatewayCommand):
    def __init__(self, id: int, actorId: int):
        super().__init__(CommeoSenSimCommand.DELETE, [(ParameterType.INT, id), (ParameterType.INT, actorId)])


class SenSimGetConfig(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenSimCommand.GETCONFIG, [(ParameterType.INT, id)])


class SenSimSetConfig(GatewayCommand):
    def __init__(self, id: int, activity: bool):
        super().__init__(CommeoSenSimCommand.SETCONFIG,
                         [(ParameterType.INT, id), (ParameterType.INT, 1 if activity else 0)])


class SenSimSetLabel(GatewayCommand):
    def __init__(self, id: bool, name: str):
        super().__init__(CommeoSenSimCommand.SETLABEL, [(ParameterType.INT, id), (ParameterType.STRING, name)])


class SenSimSetValues(GatewayCommand):
    def __init__(
            self,
            id: int,
            windDigital: int,
            rainDigital: int,
            tempDigital: int,
            lightDigital: int,
            tempAnalog: int,
            windAnalog: int,
            sun1Analog: int,
            dayLightAnalog: int,
            sun2Analog: int,
            sun3Analog: int
    ):
        super().__init__(CommeoSenSimCommand.SETVALUES,
                         [(ParameterType.INT, id),
                          (ParameterType.INT, windDigital),
                          (ParameterType.INT, rainDigital),
                          (ParameterType.INT, tempDigital),
                          (ParameterType.INT, lightDigital),
                          (ParameterType.INT, tempAnalog),
                          (ParameterType.INT, windAnalog),
                          (ParameterType.INT, sun1Analog),
                          (ParameterType.INT, dayLightAnalog),
                          (ParameterType.INT, sun2Analog),
                          (ParameterType.INT, sun3Analog)])


class SenSimGetValues(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenSimCommand.GETVALUES, [(ParameterType.INT, id)])


class SenSimGetIds(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoSenSimCommand.GETIDS)


class SenSimFactory(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenSimCommand.FACTORY, [(ParameterType.INT, id)])


class SenSimDrive(GatewayCommand):
    def __init__(self, id: int, command: SenSimCommandType):
        super().__init__(CommeoSenSimCommand.DRIVE, [(ParameterType.INT, id), (ParameterType.INT, command)])


class SenSimSetTest(GatewayCommand):
    def __init__(self, id: int, testMode: int):
        super().__init__(CommeoSenSimCommand.SETTEST, [(ParameterType.INT, id), (ParameterType.INT, testMode)])


class SenSimGetTest(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(CommeoSenSimCommand.GETTEST, [(ParameterType.INT, id)])


class SenSimStoreResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimDeleteResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimGetConfigResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = str(parameters[0][1])
        self.senSimId = int(parameters[1][1])
        self.activity = bool(parameters[2][1])


class SenSimSetConfigResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimSetLabelResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimSetValuesResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimGetValuesResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.sensorState = None
        self.windDigital = windDigital(int(parameters[1][1]))
        self.rainDigital = rainDigital(int(parameters[2][1]))
        self.tempDigital = tempDigital(int(parameters[3][1]))
        self.lightDigital = lightDigital(int(parameters[4][1]))
        self.tempAnalog = int(parameters[5][1])
        self.windAnalog = int(parameters[6][1])
        self.sun1Analog = int(parameters[7][1])
        self.dayLightAnalog = int(parameters[8][1])
        self.sun2Analog = int(parameters[9][1])
        self.sun3Analog = int(parameters[10][1])


class SenSimGetIdsResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ids = [b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[0][1]))]


class SenSimFactoryResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimDriveResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimSetTestResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class SenSimGetTestResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.id = int(parameters[0][1])
        self.testMode = bool(parameters[1][1])



from selve import SelveTypes, DeviceType, Util, CommunicationType


class SelveSender:
    def __init__(self, id: int, device_type: SelveTypes = SelveTypes.SENDER,
                 device_sub_type: DeviceType = DeviceType.UNKNOWN):
        self.id = id
        self.rfAddress = ""
        self.channel = 0
        self.resetCount = 0
        self.device_type = device_type
        self.device_sub_type = device_sub_type
        self.mask = Util.singlemask(id)
        self.name = "None"
        self.communicationType = CommunicationType.COMMEO
        self.lastEvent = None


from __future__ import annotations

import asyncio
import queue
import threading
import time
from itertools import chain
from typing import Callable

import serial
from serial.tools import list_ports
import untangle

from selve.commands import param, service
from selve.commands.device import *
from selve.commands.command import *
from selve.commands.event import *
from selve.commands.group import *
from selve.commands.iveo import *
from selve.commands.service import *
from selve.commands.param import *
from selve.commands.event import *
from selve.commands.senSim import *
from selve.commands.sensor import *
from selve.commands.sender import *
from selve.device import SelveDevice
from selve.group import SelveGroup
from selve.iveo import IveoDevice
from selve.senSim import SelveSenSim
from selve.sender import SelveSender
from selve.sensor import SelveSensor
from selve.util import *
from selve.util import Command
from selve.util.errors import *
from selve.util.protocol import ParameterType


class Selve:
    """Implementation of the serial communication to the Selve Gateway"""

    def __init__(self, port, discover=True, develop=False, logger=None):
        # Gateway state
        self._callbacks = set()
        self.lastLogEvent = None
        self.state = None

        # Data from Duty Cycle Event
        self.utilization = 0
        self.sendingBlocked = DutyMode.NOT_BLOCKED

        # Known devices
        self.devices: dict = {
            SelveTypes.DEVICE.value: {},
            SelveTypes.IVEO.value: {},
            SelveTypes.GROUP.value: {},
            SelveTypes.SENSIM.value: {},
            SelveTypes.SENSOR.value: {},
            SelveTypes.SENDER.value: {}
        }

        # Flags for enabling reader and writer in the worker thread
        self._pauseReader = False
        self._pauseWriter = False

        # Port where the Selve gateway was found
        self._port = port

        # Write lock to safely write to the gateway
        self._writeLock = threading.Lock()

        self._LOGGER = logger

        self._setup()
        self._start()

        if discover:
            self._LOGGER.info("Discovering devices")
            self.discover()

    def _setup(self):
        self._LOGGER.info("Setup")
        self.rxQ = queue.Queue()
        self.txQ = queue.Queue()

        available_ports = list_ports.comports()

        if len(available_ports) == 0:
            self._LOGGER.error("No available comports!")
            raise PortError

        for p in available_ports:
            try:
                self._serial = serial.Serial(
                    port=p.device,
                    baudrate=115200,
                    bytesize=serial.EIGHTBITS,
                    parity=serial.PARITY_NONE,
                    stopbits=serial.STOPBITS_ONE,
                    xonxoff=False,
                    rtscts=False,
                    dsrdtr=False)
                if self.pingGateway():
                    self._port = p.name
                    return
                else:
                    self._serial = None
            except PortError:
                pass
        else:
            self._LOGGER.error("No gateway on comports found!")
            raise PortError

    def register_callback(self, callback: Callable[[], None]) -> None:
        """Register callback, called when Roller changes state."""
        self._callbacks.add(callback)

    def remove_callback(self, callback: Callable[[], None]) -> None:
        """Remove previously registered callback."""
        self._callbacks.discard(callback)

    def _worker(self):
        # Infinite loop to collect all incoming data
        self._LOGGER.debug("Reader started")
        try:
            while True:
                if not self._pauseReader:
                    if self._serial.in_waiting > 0:
                        msg = ""
                        while True:
                            response = self._serial.readline().strip()
                            msg += response.decode()
                            if response.decode() == '':
                                break

                        # do something with the received data
                        self.processResponse(msg)

                        # if msg.rstrip() == b' ':
                        self._LOGGER.debug(f'Received: {msg}')

                if not self._pauseWriter:
                    if not self.txQ.empty():
                        data = self.txQ.get_nowait()

                        self._sendCommandToGateway(data)

                        self.txQ.task_done()

                        # always sleep after writing
                        time.sleep(0.5)

                time.sleep(0.01)
        # serial port exceptions, all of these notify that we are in some
        # serious trouble
        except serial.SerialException:
            # log message
            self._LOGGER.error('Serial Port RX error')

    def _sendCommandToGateway(self, command: Command):
        commandstr = command.serializeToXML()
        self._LOGGER.debug('Gateway writing: ' + str(commandstr))
        try:
            with self._writeLock:
                if not self._serial.is_open:
                    self._serial.open()
                self._serial.write(commandstr)
                self._serial.flush()
        except Exception as e:
            self._LOGGER.error("error communicating: " + str(e))
        # self.writer.close()

    async def _start(self):
        """Start all looping threads."""
        self.readLoopTask = threading.Thread(target=self._worker)
        self.readLoopTask.daemon = False
        self.readLoopTask.start()

    # close the serial port, do the cleanup
    def close(self):
        # wait for the rx/tx thread to end, these need to be gathered to
        # collect all the exceptions
        self.readLoopTask.join(5)
        self._serial.close()

    def processResponse(self, xmlstr):
        """Processes an XML String into a response object."""
        # check which command was received
        # do something with the data
        # return the ready to eat response

        # The selve device sometimes answers a badformed header. This is a patch
        xmlstr = str(xmlstr).replace('<?xml version="1.0"? encoding="UTF-8">', '<?xml version="1.0" encoding="UTF-8"?>')
        try:
            res = untangle.parse(xmlstr)
            if not hasattr(res, 'methodResponse') and not hasattr(res, 'methodCall'):
                self._LOGGER.error("Bad response format")
                return None
            if hasattr(res, 'methodResponse'):
                if hasattr(res.methodResponse, 'fault'):
                    return self.create_error(res)
                else:
                    response = self.create_response(res)
            else:
                response = self.create_response_call(res)
            # if it's a MethodResponse, it has not been sent by the gateway itself, so we can safely return it
            # otherwise it's an event, and we have to process it accordingly
            if isinstance(response, CommeoDeviceEventResponse) \
                    or isinstance(response, SensorEventResponse) \
                    or isinstance(response, SenderEventResponse) \
                    or isinstance(response, LogEventResponse) \
                    or isinstance(response, DutyCycleResponse):
                return self.processEventResponse(response)
            if isinstance(response, CommandResultResponse)\
                    or isinstance(response, IveoResultResponse):
                #update device values
                self.commandResult(response)
            if isinstance(response, DeviceGetValuesResponse):
                self.updateCommeoDeviceValuesFromResponse(int(response.parameters[1][1]), response)

            return response

        except Exception as e:
            self._LOGGER.error("Error in XML: " + str(e) + " : " + xmlstr)

    def create_error(self, obj):
        if hasattr(obj, "methodResponse"):
            return ErrorResponse(obj.methodResponse.fault.array.string.cdata, obj.methodResponse.fault.array.int.cdata)
        else:
            raise CommunicationError()

    def create_response(self, obj):
        if hasattr(obj, "methodResponse"):
            array = obj.methodResponse.array
            return self._create_response(array)
        else:
            raise CommunicationError()

    def create_response_call(self, obj):
        if hasattr(obj, "methodCall"):
            array = obj.methodCall.array
            return self._create_response(array, obj.methodCall.methodName)
        else:
            raise CommunicationError()

    def _create_response(self, array, methodName = ""):
        str_params = []
        if hasattr(array, "string"):
            if methodName == "":
                methodName = list(array.string)[0].cdata
            str_params_tmp = list(array.string)[1:]
            str_params = [(ParameterType.STRING, v.cdata) for v in str_params_tmp]
        int_params = []
        if hasattr(array, str(ParameterType.INT.value)):
            int_params = [(ParameterType.INT, v.cdata) for v in list(array.int)]
        b64_params = []
        if hasattr(array, str(ParameterType.BASE64.value)):
            b64_params = [(ParameterType.BASE64, v.cdata) for v in list(array.base64)]
        paramslist = [str_params, int_params, b64_params]
        flat_params_list = list(chain.from_iterable(paramslist))

        ##Service
        if methodName == "selve.GW." + str(CommeoServiceCommand.PING.value):
            return ServicePingResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoServiceCommand.GETSTATE.value):
            return ServiceGetStateResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoServiceCommand.GETVERSION.value):
            return ServiceGetVersionResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoServiceCommand.RESET.value):
            return ServiceResetResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoServiceCommand.FACTORYRESET.value):
            return ServiceFactoryResetResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoServiceCommand.SETLED.value):
            return ServiceSetLedResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoServiceCommand.GETLED.value):
            return ServiceGetLedResponse(methodName, flat_params_list)

        ##Param
        if methodName == "selve.GW." + str(CommeoParamCommand.SETFORWARD.value):
            return ParamSetForwardResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoParamCommand.GETFORWARD.value):
            return ParamGetForwardResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoParamCommand.SETEVENT.value):
            return ParamSetEventResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoParamCommand.GETEVENT.value):
            return ParamGetEventResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoParamCommand.GETDUTY.value):
            return ParamGetDutyResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoParamCommand.GETRF.value):
            return ParamGetRfResponse(methodName, flat_params_list)

        ##Device
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SCANSTART.value):
            return DeviceScanStartResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SCANSTOP.value):
            return DeviceScanStopResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SCANRESULT.value):
            return DeviceScanResultResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SAVE.value):
            return DeviceSaveResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.GETIDS.value):
            return DeviceGetIdsResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.GETINFO.value):
            return DeviceGetInfoResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.GETVALUES.value):
            return DeviceGetValuesResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SETFUNCTION.value):
            return DeviceSetFunctionResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SETLABEL.value):
            return DeviceSetLabelResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.SETTYPE.value):
            return DeviceSetTypeResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.DELETE.value):
            return DeviceDeleteResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoDeviceCommand.WRITEMANUAL.value):
            return DeviceWriteManualResponse(methodName, flat_params_list)

        ##Sensor
        if methodName == "selve.GW." + str(CommeoSensorCommand.TEACHSTART.value):
            return SensorTeachStartResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.TEACHSTOP.value):
            return SensorTeachStopResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.TEACHRESULT.value):
            return SensorTeachResultResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.GETIDS.value):
            return SensorGetIdsResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.GETINFO.value):
            return SensorGetInfoResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.GETVALUES.value):
            return SensorGetValuesResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.SETLABEL.value):
            return SensorSetLabelResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.DELETE.value):
            return SensorDeleteResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSensorCommand.WRITEMANUAL.value):
            return SensorWriteManualResponse(methodName, flat_params_list)

        ##SenSim
        if methodName == "selve.GW." + str(CommeoSenSimCommand.STORE.value):
            return SenSimStoreResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.DELETE.value):
            return SenSimDeleteResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.GETCONFIG.value):
            return SenSimGetConfigResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.SETCONFIG.value):
            return SenSimSetConfigResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.SETLABEL.value):
            return SenSimSetLabelResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.SETVALUES.value):
            return SenSimSetValuesResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.GETVALUES.value):
            return SenSimGetValuesResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.GETIDS.value):
            return SenSimGetIdsResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.FACTORY.value):
            return SenSimFactoryResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.DRIVE.value):
            return SenSimDriveResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.SETTEST.value):
            return SenSimSetTestResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenSimCommand.GETTEST.value):
            return SenSimGetTestResponse(methodName, flat_params_list)

        ##Sender
        if methodName == "selve.GW." + str(CommeoSenderCommand.TEACHSTART.value):
            return SenderTeachStartResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.TEACHSTOP.value):
            return SenderTeachStopResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.TEACHRESULT.value):
            return SenderTeachResultResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.GETIDS.value):
            return SenderGetIdsResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.GETINFO.value):
            return SenderGetInfoResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.GETVALUES.value):
            return SenderGetValuesResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.SETLABEL.value):
            return SenderSetLabelResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.DELETE.value):
            return SenderDeleteResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoSenderCommand.WRITEMANUAL.value):
            return SenderWriteManualResponse(methodName, flat_params_list)

        ##Group
        if methodName == "selve.GW." + str(CommeoGroupCommand.READ.value):
            return GroupReadResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoGroupCommand.WRITE.value):
            return GroupWriteResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoGroupCommand.GETIDS.value):
            return GroupGetIdsResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoGroupCommand.DELETE.value):
            return GroupDeleteResponse(methodName, flat_params_list)

        ##Command
        if methodName == "selve.GW." + str(CommeoCommandCommand.DEVICE.value):
            return CommandDeviceResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoCommandCommand.GROUP.value):
            return CommandGroupResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoCommandCommand.GROUPMAN.value):
            return CommandGroupManResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoCommandCommand.RESULT.value):
            return CommandResultResponse(methodName, flat_params_list)

        ##Iveo
        if methodName == "selve.GW." + str(IveoCommand.FACTORY.value):
            return IveoFactoryResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.SETCONFIG.value):
            return IveoSetConfigResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.GETCONFIG.value):
            return IveoGetConfigResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.GETIDS.value):
            return IveoGetIdsResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.SETREPEATER.value):
            return IveoSetRepeaterResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.GETREPEATER.value):
            return IveoGetRepeaterResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.SETLABEL.value):
            return IveoSetLabelResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.TEACH.value):
            return IveoTeachResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.LEARN.value):
            return IveoLearnResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.MANUAL.value):
            return IveoManualResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.AUTOMATIC.value):
            return IveoAutomaticResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(IveoCommand.RESULT.value):
            return IveoResultResponse(methodName, flat_params_list)

        ##Events
        if methodName == "selve.GW." + str(CommeoEventCommand.DEVICE.value):
            return CommeoDeviceEventResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoEventCommand.SENSOR.value):
            return SensorEventResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoEventCommand.SENDER.value):
            return SenderEventResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoEventCommand.LOG.value):
            return LogEventResponse(methodName, flat_params_list)
        if methodName == "selve.GW." + str(CommeoEventCommand.DUTYCYCLE.value):
            return DutyCycleResponse(methodName, flat_params_list)

        # Any other response (unknown)
        return MethodResponse(methodName, flat_params_list)

    def executeCommand(self, command: Command):
        self.txQ.put_nowait(command)

    def executeCommandSyncWithResponse(self, command: Command):

        self._pauseWriter = True
        self._pauseReader = True
        self._pauseWorker = True
        if not self._serial.is_open:
            self._serial.open()

        self._sendCommandToGateway(command)

        start_time = time.time()
        while True:
            if self._serial.in_waiting > 0:
                msg = ""
                while True:
                    response = self._serial.readline().strip()
                    msg += response.decode()
                    if response.decode() == '':
                        break
                # if msg.rstrip() == b' ':
                self._LOGGER.debug(f'Received: {msg}')

                self._pauseReader = False
                self._pauseWriter = False
                self._pauseWorker = False

                resp = self.processResponse(msg)
                if isinstance(resp, ErrorResponse):
                    self._LOGGER.error(resp.message)
                    raise GatewayError
                else:
                    #time.sleep(0.5)
                    return resp
            if time.time() - start_time > 10:
                time.sleep(0.05)
                return None


    def discover(self):
        if self.gatewayReady():
            iveoIds: IveoGetIdsResponse = self.executeCommandSyncWithResponse(IveoGetIds())
            deviceIds: DeviceGetIdsResponse = self.executeCommandSyncWithResponse(DeviceGetIds())
            groupIds: GroupGetIdsResponse = self.executeCommandSyncWithResponse(GroupGetIds())
            sensorIds: SensorGetIdsResponse = self.executeCommandSyncWithResponse(SensorGetIds())
            senderIds: SenderGetIdsResponse = self.executeCommandSyncWithResponse(SenderGetIds())
            senSimIds: SenSimGetIdsResponse = self.executeCommandSyncWithResponse(SenSimGetIds())

            for i in iveoIds.ids:
                config: IveoGetConfigResponse = self.executeCommandSyncWithResponse(IveoGetConfig(i))
                device = IveoDevice(i, device_sub_type=config.deviceType)
                device.name = config.name
                device.activity = config.activity
                self.addOrUpdateDevice(device, SelveTypes.IVEO)

            for i in deviceIds.ids:
                config: DeviceGetInfoResponse = self.executeCommandSyncWithResponse(DeviceGetInfo(i))
                device = SelveDevice(i, device_type=SelveTypes.DEVICE, device_sub_type=config.deviceType)
                device.name = config.name
                device.device_sub_type = config.deviceType
                device.rfAdress = config.rfAddress
                device.infoState = config.state
                self.addOrUpdateDevice(device, SelveTypes.DEVICE)
                config: DeviceGetValuesResponse = self.executeCommandSyncWithResponse(DeviceGetValues(i))
                device.state = config.movementState
                device.value = config.value
                device.targetValue = config.targetValue
                device.unreachable = config.unreachable
                device.overload = config.overload
                device.obstructed = config.obstructed
                device.alarm = config.alarm
                device.lostSensor = config.lostSensor
                device.automaticMode = config.automaticMode
                device.gatewayNotLearned = config.gatewayNotLearned
                device.windAlarm = config.windAlarm
                device.rainAlarm = config.rainAlarm
                device.freezingAlarm = config.freezingAlarm
                device.dayMode = config.dayMode
                self.addOrUpdateDevice(device, SelveTypes.DEVICE)

            for i in groupIds.ids:
                config: GroupReadResponse = self.executeCommandSyncWithResponse(GroupRead(i))
                device = SelveGroup(i)
                device.device_type = SelveTypes.GROUP
                device.name = config.name
                device.mask = config.mask
                self.addOrUpdateDevice(device, SelveTypes.GROUP)

            for i in sensorIds.ids:
                device = SelveSensor(i)
                config: SensorGetInfoResponse = self.executeCommandSyncWithResponse(SensorGetInfo(i))
                device.rfAdress = config.rfAddress
                device.device_type = SelveTypes.SENSOR
                self.addOrUpdateDevice(device, SelveTypes.SENSOR)
                config: SensorGetValuesResponse = self.executeCommandSyncWithResponse(SensorGetValues(i))
                device.windDigital = config.windDigital
                device.rainDigital = config.rainDigital
                device.tempDigital = config.tempDigital
                device.lightDigital = config.lightDigital
                device.sensorState = config.sensorState
                device.tempAnalog = config.tempAnalog
                device.windAnalog = config.windAnalog
                device.sun1Analog = config.sun1Analog
                device.dayLightAnalog = config.dayLightAnalog
                device.sun2Analog = config.sun2Analog
                device.sun3Analog = config.sun3Analog
                self.addOrUpdateDevice(device, SelveTypes.SENSOR)

            for i in senderIds.ids:
                config: SenderGetInfoResponse = self.executeCommandSyncWithResponse(SenderGetInfo(i))
                device = SelveSender(i)
                device.device_type = SelveTypes.SENDER
                device.name = config.name
                device.rfAdress = config.rfAddress
                device.channel = config.rfChannel
                device.resetCount = config.rfResetCount
                self.addOrUpdateDevice(device, SelveTypes.SENDER)

            for i in senSimIds.ids:
                config: SenSimGetConfigResponse = self.executeCommandSyncWithResponse(SenSimGetConfig(i))
                device = SelveSenSim(i)
                device.activity = config.activity
                device.device_type = SelveTypes.SENSIM
                self.addOrUpdateDevice(device, SelveTypes.SENSIM)
                config: SenSimGetValuesResponse = self.executeCommandSyncWithResponse(SenSimGetValues(i))
                device.windDigital = config.windDigital
                device.rainDigital = config.rainDigital
                device.tempDigital = config.tempDigital
                device.lightDigital = config.lightDigital
                device.sensorState = config.sensorState
                device.tempAnalog = config.tempAnalog
                device.windAnalog = config.windAnalog
                device.sun1Analog = config.sun1Analog
                device.dayLightAnalog = config.dayLightAnalog
                device.sun2Analog = config.sun2Analog
                device.sun3Analog = config.sun3Analog
                self.addOrUpdateDevice(device, SelveTypes.SENSIM)
        self.list_devices()


    def updateAllDevices(self):
        for device in self.devices[SelveTypes.DEVICE.value]:
            self.updateCommeoDeviceValues(device.id)
        for sensor in self.devices[SelveTypes.SENSOR.value]:
            pass
        for senSim in self.devices[SelveTypes.SENSIM.value]:
            pass
        for sender in self.devices[SelveTypes.SENDER.value]:
            pass



    def addOrUpdateDevice(self, device, type: SelveTypes):
        self.devices[type.value][device.id] = device
        # add in gateway

        # if there is a callback for updates, call it
        for callback in self._callbacks:
            callback()

    def getDevice(self, id: int, type: SelveTypes) -> SelveDevice | SelveSensor | SelveSender | SelveGroup | SelveSenSim | None:
        if id in self.devices[type.value]:
            return self.devices[type.value][id]
        return None


    def deleteDevice(self, id, type: SelveTypes):
        # delete in GW
        self.devices[type.value].pop(id)

    def is_id_registered(self, id, type: SelveTypes):
        return id in self.devices[type.value]

    def findFreeId(self, type: SelveTypes):
        i = 0
        boundary = 1
        if type is SelveTypes.SENDER:
            boundary = 62
        if type is SelveTypes.SENSOR:
            boundary = 7
        if type is SelveTypes.DEVICE:
            boundary = 63
        if type is SelveTypes.GROUP:
            boundary = 31
        if type is SelveTypes.IVEO:
            boundary = 63
        if type is SelveTypes.SENSIM:
            boundary = 7

        while i < boundary:
            if not self.is_id_registered(i, type):
                return i
            i = i + 1

    def pingGateway(self):
        cmd = service.ServicePing()
        methodResponse = self.executeCommandSyncWithResponse(cmd)
        if hasattr(methodResponse, "name"):
            if methodResponse.name == "selve.GW.service.ping":
                self._LOGGER.debug("Ping back")
                return True
        self._LOGGER.debug("No ping")
        return False


    def gatewayState(self):
        cmd = service.ServiceGetState()
        try:
            methodResponse = self.executeCommandSyncWithResponse(cmd)
        except GatewayError:
            self._LOGGER.error(str(GatewayError))
            methodResponse = None

        if hasattr(methodResponse, "name"):
            if methodResponse.name == "selve.GW." + str(CommeoServiceCommand.GETSTATE.value):
                if hasattr(methodResponse, "parameters"):
                    status = ServiceState(int(methodResponse.parameters[0][1]))
                    self._LOGGER.debug(f'Gateway state: {status}')
                    self.state = status
                    return status
        return None

    def gatewayReady(self):
        state = self.gatewayState()
        return state is ServiceState.READY

    def getVersionG(self):
        cmd = service.ServiceGetVersion()
        methodResponse = self.executeCommandSyncWithResponse(cmd)
        return methodResponse

    def getGatewayFirmwareVersion(self):
        command = self.getVersionG()
        if hasattr(command, "version"):
            return command.version
        else:
            return False

    def getGatewaySerial(self):
        command = self.getVersionG()
        if hasattr(command, "serial"):
            return command.serial
        else:
            return False

    def getGatewaySpec(self):
        command = self.getVersionG()
        if hasattr(command, "spec"):
            return command.spec
        else:
            return False

    def list_devices(self):
        """[summary]
        Log the list of registered devices
        """
        for id, val in self.devices.items():
            for id, device in val.items():
                self._LOGGER.info(str(device))

    def resetGateway(self):
        command = service.ServiceReset()
        response: ServiceResetResponse = self.executeCommandSyncWithResponse(command)
        if response.executed != 1:
            self._LOGGER.info("Error: Gateway could not be reset or loads too long")

        # time.sleep(2)

        start_time = time.time()
        while self.gatewayState() != ServiceState.READY:
            if time.time() - start_time >= 10:
                self._LOGGER.info("Error: Gateway could not be reset or loads too long")
            pass
        self._LOGGER.info("Gateway reset")

    def processEventResponse(self, response):
        if isinstance(response, CommeoDeviceEventResponse):
            # This is a commeo device response, Iveo does not generate events because it is a one way communication protocol
            if self.is_id_registered(response.id, SelveTypes.DEVICE):
                device: SelveDevice = self.devices[SelveTypes.DEVICE.value][response.id]
            else:
                device = SelveDevice(response.id, SelveTypes.DEVICE, response.deviceType)
                device.name = response.name
                device.communicationType = CommunicationType.COMMEO
                self._LOGGER.error("Id not found, creating")

            device.state = response.actorState
            device.value = response.value
            device.targetValue = response.targetValue
            device.unreachable = response.unreachable
            device.overload = response.overload
            device.obstructed = response.obstructed
            device.alarm = response.alarm
            device.lostSensor = response.lostSensor
            device.automaticMode = response.automaticMode
            device.gatewayNotLearned = response.gatewayNotLearned
            device.windAlarm = response.windAlarm
            device.rainAlarm = response.rainAlarm
            device.freezingAlarm = response.freezingAlarm
            device.dayMode = response.dayMode
            device.device_type = response.deviceType

            self.addOrUpdateDevice(device, SelveTypes.DEVICE)

        if isinstance(response, SensorEventResponse):
            if self.is_id_registered(response.id, SelveTypes.SENSOR):
                sensor: SelveSensor = self.devices[SelveTypes.SENSOR.value][response.id]
            else:
                sensor = SelveSensor(response.id)
                self._LOGGER.error("Id not found, creating")

            sensor.windDigital = response.windDigital
            sensor.rainDigital = response.rainDigital
            sensor.tempDigital = response.tempDigital
            sensor.lightDigital = response.lightDigital
            sensor.sensorState = response.sensorState
            sensor.tempAnalog = response.tempAnalog
            sensor.windAnalog = response.windAnalog
            sensor.sun1Analog = response.sun1Analog
            sensor.dayLightAnalog = response.dayLightAnalog
            sensor.sun2Analog = response.sun2Analog
            sensor.sun3Analog = response.sun3Analog
            self.addOrUpdateDevice(sensor, SelveTypes.SENSOR)

        if isinstance(response, SenderEventResponse):
            if self.is_id_registered(response.id, SelveTypes.SENDER):
                sender: SelveSender = self.getDevice(response.id, SelveTypes.SENDER)
            else:
                sender = SelveSender(response.id)
                self._LOGGER.info("Id not found, creating")

            sender.lastEvent = response.event
            sender.name = response.senderName
            self.addOrUpdateDevice(sender, SelveTypes.SENSOR)

        if isinstance(response, LogEventResponse):
            self.lastLogEvent = response
            if response.logType == LogType.INFO:
                self._LOGGER.info(
                    f'Gateway Log Info: {response.logCode} - {response.logStamp} - {response.logValue} - {response.logDescription}')
            if response.logType == LogType.WARNING:
                self._LOGGER.warning(
                    f'Gateway Log Info: {response.logCode} - {response.logStamp} - {response.logValue} - {response.logDescription}')
            if response.logType == LogType.ERROR:
                self._LOGGER.error(
                    f'Gateway Log Info: {response.logCode} - {response.logStamp} - {response.logValue} - {response.logDescription}')

        if isinstance(response, DutyCycleResponse):
            self.sendingBlocked = response.mode
            self.utilization = response.traffic


    def commandResult(self, response):
        for callback in self._callbacks:
            callback()

    ##Device functions

    def updateCommeoDeviceValues(self, id: int):
        response: DeviceGetValuesResponse = self.executeCommandSyncWithResponse(DeviceGetValues(id))
        self.updateCommeoDeviceValuesFromResponse(id, response)

    def updateCommeoDeviceValuesAsync(self, id: int):
        self.executeCommand(DeviceGetValues(id))
        
    def updateCommeoDeviceValuesFromResponse(self, id: int, response: DeviceGetValuesResponse):
        dev = self.getDevice(id, SelveTypes.DEVICE)
        dev.name = response.name
        dev.state = response.movementState
        dev.value = response.value
        dev.targetValue = response.targetValue
        dev.unreachable = response.unreachable
        dev.overload = response.overload
        dev.obstructed = response.obstructed
        dev.alarm = response.alarm
        dev.lostSensor = response.lostSensor
        dev.automaticMode = response.automaticMode
        dev.gatewayNotLearned = response.gatewayNotLearned
        dev.windAlarm = response.windAlarm
        dev.rainAlarm = response.rainAlarm
        dev.freezingAlarm = response.freezingAlarm
        dev.dayMode = response.dayMode
        self.addOrUpdateDevice(dev, SelveTypes.DEVICE)
        
    def setDeviceValue(self, id: int, value: int, type: SelveTypes):
        dev = self.getDevice(id, type)
        dev.value = value
        self.addOrUpdateDevice(dev, type)

    def setDeviceState(self, id: int, state: MovementState, type: SelveTypes):
        dev = self.getDevice(id, type)
        dev.state = state
        self.addOrUpdateDevice(dev, type)

    def moveDeviceUp(self, device: SelveDevice | IveoDevice, type=DeviceCommandType.MANUAL):
        if device.communicationType is CommunicationType.COMMEO:
            self.executeCommand(CommandDriveUp(device.id, type))
            device.state = MovementState.UP_ON
            self.addOrUpdateDevice(device, SelveTypes.DEVICE)
            self.updateCommeoDeviceValuesAsync(device.id)
        else:
            self.setDeviceState(device.id, MovementState.UP_ON, SelveTypes.IVEO)
            self.executeCommand(IveoManual(device.mask, DriveCommandIveo.UP))
            self.setDeviceState(device.id, MovementState.STOPPED_OFF, SelveTypes.IVEO)
            self.setDeviceValue(device.id, 100, SelveTypes.IVEO)

    def moveDeviceDown(self, device: SelveDevice | IveoDevice, type=DeviceCommandType.MANUAL):
        if device.communicationType is CommunicationType.COMMEO:
            self.executeCommand(CommandDriveDown(device.id, type))
            device.state = MovementState.DOWN_ON
            self.addOrUpdateDevice(device, SelveTypes.DEVICE)
            self.updateCommeoDeviceValuesAsync(device.id)
        else:
            self.setDeviceState(device.id, MovementState.DOWN_ON, SelveTypes.IVEO)
            self.executeCommand(IveoManual(device.mask, DriveCommandIveo.DOWN))
            self.setDeviceState(device.id, MovementState.STOPPED_OFF, SelveTypes.IVEO)
            self.setDeviceValue(device.id, 100, SelveTypes.IVEO)

    def moveDevicePos1(self, device: SelveDevice | IveoDevice, type=DeviceCommandType.MANUAL):
        if device.communicationType is CommunicationType.COMMEO:
            self.executeCommand(CommandDrivePos1(device.id, type))
            self.updateCommeoDeviceValuesAsync(device.id)
        else:
            self.setDeviceState(device.id, MovementState.UP_ON, SelveTypes.IVEO)
            self.executeCommand(IveoManual(device.mask, DriveCommandIveo.POS1))
            self.setDeviceState(device.id, MovementState.STOPPED_OFF, SelveTypes.IVEO)
            self.setDeviceValue(device.id, 33, SelveTypes.IVEO)

    def moveDevicePos2(self, device: SelveDevice | IveoDevice, type=DeviceCommandType.MANUAL):
        if device.communicationType is CommunicationType.COMMEO:
            self.executeCommand(CommandDrivePos2(device.id, type))
            self.updateCommeoDeviceValuesAsync(device.id)
        else:
            self.setDeviceState(device.id, MovementState.DOWN_ON, SelveTypes.IVEO)
            self.executeCommand(IveoManual(device.mask, DriveCommandIveo.POS2))
            self.setDeviceState(device.id, MovementState.STOPPED_OFF, SelveTypes.IVEO)
            self.setDeviceValue(device.id, 66, SelveTypes.IVEO)

    def moveDevicePos(self, device: SelveDevice, pos: int = 0, type=DeviceCommandType.MANUAL):
        self.executeCommand(CommandDrivePos(device.id, type, param=Util.percentageToValue(pos)))
        self.updateCommeoDeviceValuesAsync(device.id)

    def moveDeviceStepUp(self, device: SelveDevice, degrees: int = 0, type=DeviceCommandType.MANUAL):
        self.executeCommand(CommandDriveStepUp(device.id, type, param=Util.degreesToValue(degrees)))
        self.updateCommeoDeviceValuesAsync(device.id)

    def moveDeviceStepDown(self, device: SelveDevice, degrees: int = 0, type=DeviceCommandType.MANUAL):
        self.executeCommand(CommandDriveStepDown(device.id, type, param=Util.degreesToValue(degrees)))
        self.updateCommeoDeviceValuesAsync(device.id)

    def stopDevice(self, device: SelveDevice | IveoDevice, type=DeviceCommandType.MANUAL):
        if device.communicationType is CommunicationType.COMMEO:
            self.executeCommand(CommandStop(device.id, type))
            self.updateCommeoDeviceValuesAsync(device.id)
        else:
            self.executeCommand(IveoManual(device.mask, DriveCommandIveo.STOP))
            self.setDeviceState(device.id, MovementState.STOPPED_OFF, SelveTypes.IVEO)
            self.setDeviceValue(device.id, 50, SelveTypes.IVEO)

    ## Group
    def moveGroupUp(self, group: SelveGroup, type=DeviceCommandType.MANUAL):
        self.executeCommand(CommandDriveUpGroup(group.id, type))
        for id in Util.b64_mask_to_list(group.mask):
            self.updateCommeoDeviceValuesAsync(id)

    def moveGroupDown(self, group: SelveGroup, type=DeviceCommandType.MANUAL):
        self.executeCommand(CommandDriveDownGroup(group.id, type))
        for id in Util.b64bytes_to_bitlist(group.mask):
            self.updateCommeoDeviceValuesAsync(id)

    def stopGroup(self, group: SelveGroup, type=DeviceCommandType.MANUAL):
        self.executeCommand(CommandStopGroup(group.id, type))
        for id in Util.b64bytes_to_bitlist(group.mask):
            self.updateCommeoDeviceValuesAsync(id)


from selve.util import *
        
class EventDevice(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoEventCommand.DEVICE)
        
class EventSensor(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoEventCommand.SENSOR)
        
class EventSender(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoEventCommand.SENDER)
        
class EventLog(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoEventCommand.LOG)
        
class EventDutyCycle(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoEventCommand.DUTYCYCLE)


import logging
from enum import Enum

_LOGGER = logging.getLogger(__name__)
class DeviceType(Enum):
    UNKNOWN = 0
    SHUTTER = 1
    BLIND = 2
    AWNING = 3
    SWITCH = 4
    DIMMER = 5
    NIGHT_LIGHT = 6
    DRAWN_LIGHT = 7
    HEATING = 8
    COOLING = 9
    SWITCHDAY = 10
    GATEWAY = 11

class SelveTypes(Enum):
    SERVICE = "service"
    PARAM = "param"
    DEVICE = "device"
    SENSOR = "sensor"
    SENSIM = "senSim"
    SENDER = "sender"
    GROUP = "group"
    COMMAND = "command"
    EVENT = "event"
    IVEO = "iveo"
    COMMEO = "commeo"
    FIRMWARE = "firmware"
    UNKNOWN = "unknown"

class DeviceState(Enum):
    UNUSED = 0
    USED = 1
    TEMPORARY = 2
    STALLED = 3

class MovementState(Enum):
    UNKOWN = 0
    STOPPED_OFF = 1
    UP_ON = 2
    DOWN_ON = 3

class CommunicationType(Enum):
    COMMEO = 0
    IVEO = 1
    UNKNOWN = 99

class SenSimCommandType(Enum):
    STOP = 0
    DRIVEUP = 1
    DRIVEDOWN = 2
    POSITION_1 = 3
    POSITION_2 = 4

class DriveCommandCommeo(Enum):
    STOP = 0
    DRIVEUP = 1
    DRIVEDOWN = 2
    DRIVEPOS1 = 3
    SAVEPOS1 = 4
    DRIVEPOS2 = 5
    SAVEPOS2 = 6
    DRIVEPOS = 7
    STEPUP = 8
    STEPDOWN = 9 
    AUTOON = 10
    AUTOOFF = 11

class DeviceCommandType(Enum):
    FORCED = 0
    MANUAL = 1
    TIME = 2
    GLASS = 3

class DriveCommandIveo(Enum):
    STOP = 0
    UP = 1
    DOWN = 2
    POS1 = 3
    POS2 = 5
    LEARNTELEGRAMSENT = 254
    TEACHTELEGRAMSENT = 255

class ParameterType(Enum):
    INT = "int"
    STRING = "string"
    BASE64 = "base64"

class ScanState(Enum):
    IDLE = 0
    RUN = 1
    VERIFY = 2
    END_SUCCESS = 3
    END_FAILED = 4

class TeachState(Enum):
    IDLE = 0
    RUN = 1
    END_SUCCESS = 2

class CommandResultState(Enum):
    IDLE = 0
    SEND = 1

class ServiceState(Enum):
    BOOTLOADER = 0
    UPDATE = 1
    STARTUP = 2
    READY = 3

class SensorState(Enum):
    INVALID = 0
    AVAILABLE = 1
    LOW_BATTERY = 2
    COMMUNICATION_LOSS = 3
    TESTMODE = 4
    SERVICEMODE = 5

class RepeaterState(Enum):
    NONE = 0
    SINGLEREPEAT = 1
    MULTIREPEAT = 2

class LEDMode(Enum):
    OFF = 0
    ON = 1

class Forwarding(Enum):
    OFF = 0
    ON = 1

class DutyMode(Enum):
    NOT_BLOCKED = 0
    BLOCKED = 1

class DayMode(Enum):
    UNKOWN = 0
    NIGHTMODE = 1
    DAWNING = 2
    DAY = 3
    DUSK = 4

class DeviceFunctions(Enum):
    SELECT = 0
    INSTALL = 1
    SENSOR = 2
    MANPROG = 3
    AUTOPROG = 4
    STOREPOSITION = 5
    DRIVEUP = 6
    DRIVEDOWN = 7
    KEYRELEASE = 8
    DRIVESTOP = 9

class LogType(Enum):
    INFO = 0
    WARNING = 1
    ERROR = 2

class DeviceClass(Enum):
    ACTOR = 0
    GROUP = 1
    SENDER = 2
    SENSIM = 3
    SENSOR = 4
    IVEO = 5
    UNKOWN = 99

## SensorVariables ##

class windDigital(Enum):
    NONE = 0
    NO_ALARM = 1
    ALARM = 2

class rainDigital(Enum):
    NONE = 0
    NO_ALARM = 1
    ALARM = 2

class tempDigital(Enum):
    NONE = 0
    NORMAL = 1
    FREEZING = 2
    HEAT = 3

class lightDigital(Enum):
    NONE = 0
    DARK = 1
    DAWN = 2
    NORMAL = 3
    LIGHT = 4

## senderEvents ##

class senderEvents(Enum):
    UNKNOWN = 0
    DRIVEUP = 1
    DRIVEDOWN = 2
    STOP = 3
    POS1 = 4
    POS2 = 5
    SAVEPOS1 = 6
    SAVEPOS2 = 7
    AUTO = 8
    MAN = 9
    NAME = 10
    KEYRELEASE = 11
    SELECT = 12
    DELETE = 13





from selve.util import *


class ServicePing(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoServiceCommand.PING)


class ServiceGetState(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoServiceCommand.GETSTATE)


class ServiceGetVersion(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoServiceCommand.GETVERSION)


class ServiceReset(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoServiceCommand.RESET)


class ServiceFactoryReset(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoServiceCommand.FACTORYRESET)


class ServiceSetLed(GatewayCommand):
    def __init__(self, on: bool):
        super().__init__(CommeoServiceCommand.SETLED, [(ParameterType.INT, 1 if on else 0)])


class ServiceGetLed(GatewayCommand):
    def __init__(self):
        super().__init__(CommeoServiceCommand.GETLED)


class ServicePingResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)


class ServiceGetStateResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.state = parameters[0][1]


class ServiceGetVersionResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.serial = str(parameters[0][1])
        self.version = int(parameters[1][1]) + "." + int(parameters[2][1]) + "." + int(parameters[3][1]) + "." + int(
            parameters[6][1])
        self.spec = int(parameters[4][1]) + "." + int(parameters[5][1])


class ServiceResetResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class ServiceFactoryResetResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class ServiceSetLedResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class ServiceGetLedResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ledmode = LEDMode(int(parameters[0][1]))



import base64
from selve.util.protocol import *

class Command():
    def __init__(self, method_name, parameters = []) -> None:
        self.method_name = method_name
        self.parameters = parameters

    def serializeToXML(self):
            xmlstr = "<methodCall>"
            xmlstr += "<methodName>"+self.method_name+"</methodName>"
            if (len(self.parameters) > 0):
                xmlstr += "<array>"
                for typ, val in self.parameters:
                    xmlstr+="<{0}>{1}</{0}>".format(typ.value, val)
                xmlstr += "</array>"
            xmlstr+= "</methodCall>"
            return xmlstr.encode('utf-8')
 

class GatewayCommand(Command):

    def __init__(self, method_name, parameters = []):
         super().__init__("selve.GW." + method_name.value, parameters)

class CommandSingle(Command):

    def __init__(self, method_name, iveoID):
        super().__init__(method_name, [(ParameterType.INT, iveoID)])

class CommandMask(Command):

    def __init__(self, method_name, mask, command):
        super().__init__(method_name, [(ParameterType.BASE64, mask), (ParameterType.INT, command.value)])


class MethodResponse:

    def __init__(self, name, parameters):
        self.name = name
        self.parameters = parameters
        
class CommeoCommandResult(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.command = self.name
        self.commandType = DeviceCommandType(int(parameters[1][1]))
        self.executed = bool(parameters[2][1])
        self.successIds = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[3][1]))]
        self.failedIds = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[4][1]))]

class CommeoDeviceEventResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = str(parameters[0][1])
        self.id = int(parameters[1][1])
        self.actorState = MovementState(int(parameters[2][1]))
        self.value = Util.valueToPercentage(int(parameters[3][1]))
        self.targetValue = Util.valueToPercentage(int(parameters[4][1]))
        bArr = Util.intToBoolarray(int(parameters[5][1]))
        self.unreachable = bArr[0]
        self.overload = bArr[1]
        self.obstructed = bArr[2]
        self.alarm = bArr[3]
        self.lostSensor = bArr[4]
        self.automaticMode = bArr[5]
        self.gatewayNotLearned = bArr[6]
        self.windAlarm = bArr[7]
        self.rainAlarm = bArr[8]
        self.freezingAlarm = bArr[9]
        self.dayMode = DayMode(int(parameters[6][1]))
        self.deviceType = DeviceType(int(parameters[7][1]))

class LogEventResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.logCode = str(parameters[0][1])
        self.logStamp = str(parameters[1][1])
        self.logValue = str(parameters[2][1])
        self.logDescription = str(parameters[3][1])
        self.logType = LogType(int(parameters[4][1]))

class DutyCycleResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.mode = DutyMode(int(parameters[0][1]))
        self.traffic = int(parameters[1][1])

class SensorEventResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.id = int(parameters[0][1])
        self.windDigital = windDigital(int(parameters[1][1]))
        self.rainDigital = rainDigital(int(parameters[2][1]))
        self.tempDigital = tempDigital(int(parameters[3][1]))
        self.lightDigital = lightDigital(int(parameters[4][1]))
        self.sensorState = SensorState(int(parameters[5][1]))
        self.tempAnalog = int(parameters[6][1])
        self.windAnalog = int(parameters[7][1])
        self.sun1Analog = int(parameters[8][1])
        self.dayLightAnalog = int(parameters[9][1])
        self.sun2Analog = int(parameters[10][1])
        self.sun3Analog = int(parameters[11][1])

class SenderEventResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.senderName = str(parameters[0][1])
        self.id = int(parameters[1][1])
        self.event = senderEvents(int(parameters[2][1]))

class ErrorResponse:
    def __init__(self, message, code):
        self.message = message
        self.code = code


class Util():
    @classmethod
    def singlemask(self, id):
        #Obtains a base64 encoded to modify just one index
        mask =  64 * [0]
        #need to transform the position
        newid = int((id // 8) * 8  + 7 - (id % 8))    
        mask[newid] = 1
        bitstring = "".join(str(x) for x in mask)
        return base64.b64encode(self.bitstring_to_bytes(bitstring)).decode('utf8')

    @classmethod
    def multimask(self, ids):
        mask = 64 * [0]
        for id in ids:
            newid = int((id // 8) * 8  + 7 - (id % 8))    
            mask[newid] = 1
        bitstring = "".join(str(x) for x in mask)
        return base64.b64encode(self.bitstring_to_bytes(bitstring)).decode('utf8')

    @classmethod
    def bitstring_to_bytes(self, s):
        return int(s, 2).to_bytes(len(s) // 8, byteorder='big')

    @classmethod
    def b64bytes_to_bitlist(self, b):
        byts = base64.b64decode(b)
        return [bool(int(value)) for value in list(''.join([bin(by).lstrip('0b').zfill(8)[::-1] for by in byts]))]

    @classmethod
    def b64_mask_to_list(self, b):
        return self.true_in_list(self.b64bytes_to_bitlist(b))


    @classmethod
    def true_in_list(self, l):
        return [i for i,v in enumerate(l) if v]

    @classmethod
    def valueToPercentage(self, value):
        return int((value / 65535)*100)
    @classmethod
    def valueToDegrees(self, value):
        return int((value / 65535)*360)

    @classmethod
    def percentageToValue(self, perc):
        return int((65535/100)*(perc))
    @classmethod
    def degreesToValue(self, perc):
        return int((65535/360)*(perc))

    @classmethod
    def intToBoolarray(self, value):
        return [bool(bit) for bit in '{0:10b}'.format(value)]

from enum import Enum

class CommeoServiceCommand(Enum):
    PING = "service.ping"
    GETSTATE = "service.getState"
    GETVERSION = "service.getVersion"
    RESET = "service.reset"
    FACTORYRESET = "service.factoryReset"
    SETLED = "service.setLED"
    GETLED = "service.getLED"
    
class CommeoParamCommand(Enum):
    SETFORWARD = "param.setForward"
    GETFORWARD = "param.getForward"
    SETEVENT = "param.setEvent"
    GETEVENT = "param.getEvent"
    GETDUTY = "param.getDuty"
    GETRF = "param.getRF"

class CommeoDeviceCommand(Enum):
    SCANSTART = "device.scanStart"
    SCANSTOP = "device.scanStop"
    SCANRESULT = "device.scanResult"
    SAVE = "device.save"
    GETIDS = "device.getIDs"
    GETINFO = "device.getInfo"
    GETVALUES = "device.getValues"
    SETFUNCTION = "device.setFunction"
    SETLABEL = "device.setLabel"
    SETTYPE = "device.setType"
    DELETE = "device.delete"
    WRITEMANUAL = "device.writeManual"
    
class CommeoSensorCommand(Enum):
    TEACHSTART = "sensor.teachStart"
    TEACHSTOP = "sensor.teachStop"
    TEACHRESULT = "sensor.teachResult"
    GETIDS = "sensor.getIDs"
    GETINFO = "sensor.getInfo"
    GETVALUES = "sensor.getValues"
    SETLABEL = "sensor.setLabel"
    DELETE = "sensor.delete"
    WRITEMANUAL = "sensor.writeManual"

class CommeoSenSimCommand(Enum):
    STORE = "senSim.store"
    DELETE = "senSim.delete"
    GETCONFIG = "senSim.getConfig"
    SETCONFIG = "senSim.setConfig"
    SETLABEL = "senSim.setLabel"
    SETVALUES = "senSim.setValues"
    GETVALUES = "senSim.getValues"
    GETIDS = "senSim.getIDs"
    FACTORY = "senSim.factory"
    DRIVE = "senSim.drive"
    SETTEST = "senSim.setTest"
    GETTEST = "senSim.getTest"
    
class CommeoSenderCommand(Enum):
    TEACHSTART = "sender.teachStart"
    TEACHSTOP = "sender.teachStop"
    TEACHRESULT = "sender.teachResult"
    GETIDS = "sender.getIDs"
    GETINFO = "sender.getInfo"
    GETVALUES = "sender.getValues"
    SETLABEL = "sender.setLabel"
    DELETE = "sender.delete"
    WRITEMANUAL = "sender.writeManual"

class CommeoGroupCommand(Enum):
    READ = "group.read"
    WRITE = "group.write"
    GETIDS = "group.getIDs"
    DELETE = "group.delete"

class CommeoCommandCommand(Enum):
    DEVICE = "command.device"
    GROUP = "command.group"
    GROUPMAN = "command.groupMan"
    RESULT = "command.result"
    
class CommeoEventCommand(Enum):
    DEVICE = "event.device"
    SENSOR = "event.sensor"
    SENDER = "event.sender"
    LOG = "event.log"
    DUTYCYCLE = "event.dutyCycle"

class IveoCommand(Enum):
    FACTORY = "iveo.factory"
    SETCONFIG = "iveo.setConfig"
    GETCONFIG = "iveo.getConfig"
    GETIDS = "iveo.getIDs"
    SETREPEATER = "iveo.setRepeater"
    GETREPEATER = "iveo.getRepeater"
    SETLABEL = "iveo.setLabel"
    TEACH = "iveo.commandTeach"
    LEARN = "iveo.commandLearn"
    MANUAL = "iveo.commandManual"
    AUTOMATIC = "iveo.commandAutomatic"
    RESULT = "iveo.commandResult"
    

class CommandType(Enum):
    def __getattr__(self, item):
        if item != '_value_':
            return getattr(self.value, item).value
        raise AttributeError
    SERVICE = CommeoServiceCommand   
    PARAM = CommeoParamCommand
    DEVICE = CommeoDeviceCommand
    SENSOR = CommeoSensorCommand
    SENSIM = CommeoSenSimCommand
    SENDER = CommeoSenderCommand
    GROUP = CommeoGroupCommand
    COMMAND = CommeoCommandCommand
    EVENT = CommeoEventCommand
    IVEO = IveoCommand

### Responses
    
class ResponseTypes(Enum):
    COMMANDRESULT = "selve.gw.command.result"


from selve.util import *


class IveoFactory(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(IveoCommand.FACTORY, [(ParameterType.INT, id)])


class IveoSetConfig(GatewayCommand):
    def __init__(self, id: int, activity: int, type: DeviceType):
        super().__init__(IveoCommand.SETCONFIG,
                         [(ParameterType.INT, id), (ParameterType.INT, activity), (ParameterType.INT, type.value)])


class IveoGetConfig(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(IveoCommand.GETCONFIG, [(ParameterType.INT, id)])


class IveoGetIds(GatewayCommand):
    def __init__(self):
        super().__init__(IveoCommand.GETIDS)


class IveoSetRepeater(GatewayCommand):
    def __init__(self, repeaterInstalled: int):
        super().__init__(IveoCommand.SETREPEATER, [(ParameterType.INT, repeaterInstalled)])


class IveoGetRepeater(GatewayCommand):
    def __init__(self):
        super().__init__(IveoCommand.GETREPEATER)


class IveoSetLabel(GatewayCommand):
    def __init__(self, id: bool, name: str):
        super().__init__(IveoCommand.SETLABEL, [(ParameterType.INT, id), (ParameterType.STRING, name)])


class IveoTeach(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(IveoCommand.TEACH, [(ParameterType.INT, id)])


class IveoLearn(GatewayCommand):
    def __init__(self, id: int):
        super().__init__(IveoCommand.LEARN, [(ParameterType.INT, id)])


class IveoManual(GatewayCommand):
    def __init__(self, actorIds: dict, command: DriveCommandIveo):
        super().__init__(IveoCommand.MANUAL,
                         [(ParameterType.BASE64, Util.multimask(actorIds)), (ParameterType.INT, command.value)])


class IveoAutomatic(GatewayCommand):
    def __init__(self, actorIds: dict, command: DriveCommandIveo):
        super().__init__(IveoCommand.AUTOMATIC,
                         [(ParameterType.BASE64, Util.multimask(actorIds)), (ParameterType.INT, command.value)])


class IveoResult(GatewayCommand):
    def __init__(self):
        super().__init__(IveoCommand.RESULT)


class IveoFactoryResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoSetConfigResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoGetConfigResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.name = parameters[0][1]
        self.activity = parameters[2][1]
        self.deviceType = DeviceType(int(parameters[3][1]))


class IveoGetIdsResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.ids = [b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[0][1]))]


class IveoSetRepeaterResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoGetRepeaterResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.repeaterState = RepeaterState(int(parameters[0][1]))


class IveoSetLabelResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoTeachResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoLearnResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoManualResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoAutomaticResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])


class IveoResultResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.command = DriveCommandIveo(int(parameters[0][1]))
        self.state = CommandResultState(int(parameters[2][1]))
        self.executedIds = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[1][1]))]



from __future__ import annotations

from selve.util import *

class CommeoCommand(GatewayCommand):
    def __init__(self, id: int | dict, driveComm: DriveCommandCommeo, type: DeviceCommandType, param: int = 0, isGroupCmd: bool = False):
        if not isGroupCmd:
            comm = CommeoCommandCommand.DEVICE
        else:
            comm = CommeoCommandCommand.GROUP
        super().__init__(comm, [(ParameterType.INT, id), (ParameterType.INT, driveComm.value),
                                (ParameterType.INT, type.value), (ParameterType.INT, param)])

class CommandStop(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.STOP, type, param, False)

class CommandDriveUp(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEUP, type, param, False)

class CommandDriveDown(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEDOWN, type, param, False)

class CommandDrivePos1(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEPOS1, type, param, False)

class CommandSavePos1(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.SAVEPOS1, type, param, False)

class CommandDrivePos2(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEPOS2, type, param, False)

class CommandSavePos2(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.SAVEPOS2, type, param, False)

class CommandDrivePos(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEPOS, type, param, False)

class CommandDriveStepUp(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.STEPUP, type, param, False)

class CommandDriveStepDown(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.STEPDOWN, type, param, False)

class CommandDriveAutoOn(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.AUTOON, type, param, False)

class CommandDriveAutoOff(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.AUTOOFF, type, param, False)


class CommandStopGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.STOP, type, param, True)

class CommandDriveUpGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEUP, type, param, True)

class CommandDriveDownGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEDOWN, type, param, True)

class CommandDrivePos1Group(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEPOS1, type, param, True)

class CommandSavePos1Group(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.SAVEPOS1, type, param, True)

class CommandDrivePos2Group(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEPOS2, type, param, True)

class CommandSavePos2Group(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.SAVEPOS2, type, param, True)

class CommandDrivePosGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.DRIVEPOS, type, param, True)

class CommandStepUpGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.STEPUP, type, param, True)

class CommandStepDownGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.STEPDOWN, type, param, True)

class CommandAutoOnGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.AUTOON, type, param, True)

class CommandAutoOffGroup(CommeoCommand):
    def __init__(self, id: int, type: DeviceCommandType, param: int = 0):
        super().__init__(id, DriveCommandCommeo.AUTOOFF, type, param, True)



class CommandGroupMan(GatewayCommand):
    def __init__(self, command, type: DeviceCommandType, ids: dict, param: int = 0):
        super().__init__(CommeoCommandCommand.GROUPMAN, [(ParameterType.INT, command.value), (ParameterType.INT, type.value), (ParameterType.BASE64, Util.multimask(ids)), (ParameterType.INT, param)])



class CommandDeviceResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class CommandGroupResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])

class CommandGroupManResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.executed = bool(parameters[0][1])
        self.ids = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[1][1]))]


class CommandResultResponse(MethodResponse):
    def __init__(self, name, parameters):
        super().__init__(name, parameters)
        self.command = DriveCommandIveo(int(parameters[0][1]))
        self.commandType = DeviceCommandType(int(parameters[1][1]))
        self.executed = bool(parameters[2][1])
        self.successIds = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[3][1]))]
        self.failedIds = [ b for b in Util.true_in_list(Util.b64bytes_to_bitlist(parameters[4][1]))]


from enum import Enum
from typing import Dict


class GatewayError(Exception):
    pass

class PortError(Exception):
    pass
class CommunicationError(Exception):
    pass
class ReadTimeoutError(Exception):
    pass


ErrorCodes = {
    1: "Unknown Error",
    2: "Method not supported",
    3: "Method not reachable",
    4: "Parameter count",
    5: "Parameter order",
    6: "Execution failed",
    7: "Parameter out of range",
    8: "Syntax error",
    9: "Method length too large",
    10: "ID is not used",
    11: "ID already exists",
    12: "Address is already used",
    13: "No member available",
    14: "Duty Cycle is reached",
    40: "Bootl: Method not supported",
    41: "Bootl: Wrong file",
    42: "Bootl: Checksum error",
    43: "Bootl: Syntax error"
}

